//Slicing of tainted data

//________________________________________________________________________________________________________________


//Library of functions
#include <idc.idc>

//________________________________________________________________________________________________________________


//Constants
#define	true		1
#define	false		0
#define	Tainted		true
#define	Untainted	false
#define	Unsupported	2
#define	RegName		0					//Offset to a register name
#define	RegOffset	1					//Offset to a value within the register array
#define	RegSize		2					//Offset to a register size
#define	RegObjectSize	3					//Size of the registry object: (name, offset, size)
#define	TaintedAddress	0x8000000000000000			//Mark of a tainted address
#define	TaintColor	0xa6998c
#define	TaintColorErr	0x0000ff
#define	TaintColorUnsup	0xd8d0c8

//Command codes (allins.hpp)
//386
#define NN_aaa		1
#define NN_aad		2
#define NN_aam		3
#define NN_aas		4
#define NN_adc		5
#define NN_add		6
#define NN_and		7
#define NN_bound	9
#define NN_bsf		10
#define NN_bsr		11
#define NN_bt		12
#define NN_btc		13
#define NN_btr		14
#define NN_bts		15
#define	NN_call		16
#define	NN_callfi	17
#define	NN_callni	18
#define	NN_cbw		19
#define	NN_cwde		20
#define	NN_cdqe		21
#define	NN_clc		22
#define	NN_cld		23
#define	NN_cmc		26
#define	NN_cmp		27
#define NN_cmps		28
#define NN_cwd		29
#define NN_cdq		30
#define NN_cqo		31
#define NN_daa		32
#define NN_das		33
#define NN_dec		34
#define NN_div		35
#define NN_enterw	36
#define NN_enter	37
#define NN_enterd	38
#define NN_enterq	39
#define NN_idiv		41
#define NN_imul		42
#define NN_inc		44
#define NN_ja		53
#define NN_jae		54
#define NN_jb		55
#define NN_jbe		56
#define NN_jc		57
#define NN_jcxz		58
#define NN_jecxz	59
#define NN_jrcxz	60
#define NN_je		61
#define NN_jg		62
#define NN_jge		63
#define NN_jl		64
#define NN_jle		65
#define NN_jna		66
#define NN_jnae		67
#define NN_jnb		68
#define NN_jnbe		69
#define NN_jnc		70
#define NN_jne		71
#define NN_jng		72
#define NN_jnge		73
#define NN_jnl		74
#define NN_jnle		75
#define NN_jno		76
#define NN_jnp		77
#define NN_jns		78
#define NN_jnz		79
#define NN_jo		80
#define NN_jp		81
#define NN_jpe		82
#define NN_jpo		83
#define NN_js		84
#define NN_jz		85
#define NN_jmp		86
#define NN_jmpfi	87
#define NN_jmpni	88
#define NN_jmpshort	89
#define	NN_lahf		90
#define	NN_lea		92
#define NN_leavew	93
#define NN_leave	94
#define NN_leaved	95
#define NN_leaveq	96
#define	NN_lgs		99
#define	NN_lss		100
#define	NN_lds		101
#define	NN_les		102
#define	NN_lfs		103
#define NN_lods		107
#define NN_loopw	108
#define NN_loop		109
#define NN_loopd	110
#define NN_loopq	111
#define NN_loopwe	112
#define NN_loope	113
#define NN_loopde	114
#define NN_loopqe	115
#define NN_loopwne	116
#define NN_loopne	117
#define NN_loopdne	118
#define NN_loopqne	119
#define	NN_mov		122
#define	NN_movs		124
#define	NN_movsx	125
#define	NN_movzx	126
#define	NN_mul		127
#define NN_neg		128
#define NN_nop		129
#define NN_not		130
#define	NN_or		131
#define	NN_pop		134
#define	NN_popaw	135
#define	NN_popa		136
#define	NN_popad	137
#define	NN_popfw	139
#define	NN_popfw	140
#define	NN_popfd	141
#define	NN_popfq	142
#define	NN_push		143
#define	NN_pushaw	144
#define NN_pusha	145
#define NN_pushad	146
#define NN_pushfw	148
#define NN_pushf	149
#define NN_pushfd	150
#define NN_pushfq	151
#define	NN_rcl		152
#define	NN_rcr		153
#define	NN_rol		154
#define	NN_ror		155
#define	NN_rep		156
#define NN_retn		159
#define NN_sahf		161
#define	NN_sal		162
#define	NN_sar		163
#define	NN_shl		164
#define	NN_shr		165
#define NN_sbb		166
#define NN_scas		167
#define NN_seta		168
#define NN_setae	169
#define NN_setb		170
#define NN_setbe	171
#define NN_setc		172
#define NN_sete		173
#define NN_setg		174
#define NN_setge	175
#define NN_setl		176
#define NN_setle	177
#define NN_setna	178
#define NN_setnae	179
#define NN_setnb	180
#define NN_setnbe	181
#define NN_setnc	182
#define NN_setne	183
#define NN_setng	184
#define NN_setnge	185
#define NN_setnl	186
#define NN_setnle	187
#define NN_setno	188
#define NN_setnp	189
#define NN_setns	190
#define NN_setnz	191
#define NN_seto		192
#define NN_setp		193
#define NN_setpe	194
#define NN_setpo	195
#define NN_sets		196
#define NN_setz		197
#define NN_shld		200
#define NN_shrd		201
#define NN_smsw		203
#define NN_stc		204
#define NN_std		205
#define NN_stos		207
#define NN_sub		209
#define NN_test		210
#define NN_xchg		214
#define NN_xor		216
#define NN_repe		242
#define NN_repne	243

//Pentium
#define NN_rdtsc	227

//Pentium Pro
#define NN_cmova	229
#define NN_cmovb	230
#define NN_cmovbe	231
#define NN_cmovg	232
#define NN_cmovge	233
#define NN_cmovl	234
#define NN_cmovle	235
#define NN_cmovnb	236
#define NN_cmovno	237
#define NN_cmovnp	238
#define NN_cmovns	239
#define NN_cmovnz	240
#define NN_cmovo	241
#define NN_cmovp	242
#define NN_cmovs	243
#define NN_cmovz	244

//Registers' numbers (intel.hpp)
#define	NN_ax		0x0
#define	NN_cx		0x1
#define	NN_dx		0x2
#define	NN_bx		0x3
#define	NN_sp		0x4
#define	NN_bp		0x5
#define	NN_si		0x6
#define	NN_di		0x7
#define	NN_r8		0x8
#define	NN_r9		0x9
#define	NN_r10		0xa
#define	NN_r11		0xb
#define	NN_r12		0xc
#define	NN_r13		0xd
#define	NN_r14		0xe
#define	NN_r15		0xf
#define	NN_al		0x10
#define	NN_cl		0x11
#define	NN_dl		0x12
#define	NN_bl		0x13
#define	NN_ah		0x14
#define	NN_ch		0x15
#define	NN_dh		0x16
#define	NN_bh		0x17
#define	NN_spl		0x18
#define	NN_bpl		0x19
#define	NN_sil		0x1a
#define	NN_dil		0x1b
#define NN_TMP		0x1c
#define NN_all		0x10					//Total number of basic registers

//Define the global array
extern $TaintMemArray;
extern $TaintMemArraySize;
extern $TaintRegArray;
extern $Registers;
extern $Bitness;
extern $Segment;
extern $CalleeFunc;
extern $TracingStop;

//________________________________________________________________________________________________________________


//Calculate a pointer of an operand and test registers for tainted
static GetPointerAddress($Op, $TaintedAddress)
{
  //Define variables
  auto $i, $addr, $BaseReg, $IdxReg, $Scale;

  //Mode: o_mem
  if ($Op.type == o_mem)
  {
    //mov eax,[401000h]
    $addr = $Op.addr;
    if ($Op.specflag1)
    {
      //mov eax,[eax*4+401000h]
      $IdxReg  = ($Op.specflag2 >> 3) & 0x7;
      $Scale   = 1 << (($Op.specflag2 >> 6) & 0x3);
      //Calculate an address
      $addr = GetTevRegVal(TEV_NONE, RegGetName($IdxReg)) * $Scale + $addr;
      //Test for tainted any register
      for ($i=0; $i<RegGetSize($IdxReg); $i++)
	if (RegCheck($IdxReg, $i))
	  return ($addr | TaintedAddress);
    }
    //Exit
    return ($addr * $TaintedAddress);
  }

  //Mode: o_phrase
  if ($Op.type == o_phrase)
  {
    if (!$Op.specflag1)
    {
      //mov eax,[eax]
      //Calculate an address
      $addr = GetTevRegVal(TEV_NONE, RegGetName($Op.reg));
      //Test for tainted any register
      for ($i=0; $i<RegGetSize($Op.reg); $i++)
	if (RegCheck($Op.reg, $i))
	  return ($addr | TaintedAddress);
    }
    else
    {
      //mov eax,[ebx+eax*4]
      $BaseReg = $Op.specflag2 & 0x7;
      $IdxReg  = ($Op.specflag2 >> 3) & 0x7;
      $Scale   = 1 << (($Op.specflag2 >> 6) & 0x3);
      //Calculate an address
      if ($BaseReg == NN_sp)
        $addr = GetTevRegVal(TEV_NONE, RegGetName($BaseReg));
      else
        $addr = GetTevRegVal(TEV_NONE, RegGetName($BaseReg)) + GetTevRegVal(TEV_NONE, RegGetName($IdxReg)) * $Scale;
      //Test for tainted any register
      for ($i=0; $i<RegGetSize($BaseReg); $i++)
	if ((RegCheck($BaseReg, $i)) || (RegCheck($IdxReg, $i)))
	  return ($addr | TaintedAddress);
    }
    //Exit
    return ($addr * $TaintedAddress);
  }

  //Mode: o_displ
  if ($Op.type == o_displ)
  {
    $addr = $Op.addr;
    if (!$Op.specflag1)
    {
      //mov eax,[eax+402031h]
      //Calculate an address
      $addr = GetTevRegVal(TEV_NONE, RegGetName($Op.reg)) + $addr;
      //Test for tainted any register
      for ($i=0; $i<RegGetSize($Op.reg); $i++)
        if (RegCheck($Op.reg, $i))
	  return ($addr | TaintedAddress);
    }
    else
    {
      //mov eax,[ebx+eax*4+402031h]
      $BaseReg = $Op.specflag2 & 0x7;
      $IdxReg  = ($Op.specflag2 >> 3) & 0x7;
      $Scale   = 1 << (($Op.specflag2 >> 6) & 0x3);
      //Calculate an address
      if ($BaseReg == NN_sp)
        $addr = GetTevRegVal(TEV_NONE, RegGetName($BaseReg)) + $addr;
      else
        $addr = GetTevRegVal(TEV_NONE, RegGetName($BaseReg)) + GetTevRegVal(TEV_NONE, RegGetName($IdxReg)) * $Scale + $addr;
      //Test for tainted any registers
      for ($i=0; $i<RegGetSize($BaseReg); $i++)
        if ((RegCheck($BaseReg, $i)) || (RegCheck($IdxReg, $i)))
	  return ($addr | TaintedAddress);
    }
    //Exit
    return ($addr * $TaintedAddress);
  }

  //Error
  throw 1;
}

//________________________________________________________________________________________________________________


//Taint the memory
static MemTaint($addr, $k)
{
  //Define variables
  auto $a = $addr & ~TaintedAddress;
  //Add the adderss to the array
  if (!MemCheck($a, $k))
  {
    $TaintMemArray[$TaintMemArraySize++] = $a + $k;
    SetColor($a + $k, CIC_ITEM, TaintColor);
  }
  //Exit
  return Tainted;
}

//________________________________________________________________________________________________________________


//Untaint the memory
static MemUntaint($addr, $k)
{
  //Define variables
  auto $i;
  auto $a = ($addr & ~TaintedAddress) + $k;
  //Checks for a tainted address
  if ($addr & TaintedAddress)
    return MemTaint($addr, $k);
  else
    for ($i=0; $i<$TaintMemArraySize; $i++)
      if ($TaintMemArray[$i] == $a)
      {
        $TaintMemArray[$i] = Untainted;
        SetColor($a, CIC_ITEM, DEFCOLOR);
      }
  //Exit
  return Untainted;
}

//________________________________________________________________________________________________________________


//Check for the already tainted memory byte
static MemCheck($addr, $k)
{
  //Define variables
  auto $i;
  auto $a = ($addr & ~TaintedAddress) + $k;
  //Checks for the tainted value
  for ($i=0; $i<$TaintMemArraySize; $i++)
    if ($TaintMemArray[$i] == $a) return Tainted;
  //Checks for a tainted address
  return ($addr & TaintedAddress) ? Tainted : Untainted;
}

//________________________________________________________________________________________________________________


//Get a registry name
static RegGetName($reg)
{
  if ($reg < NN_r8)
    return ($Bitness == 1) ? "e"+$Registers[$reg*RegObjectSize+RegName] : "r"+$Registers[$reg*RegObjectSize+RegName];
  else
    return $Registers[$reg*RegObjectSize+RegName];
}

//________________________________________________________________________________________________________________


//Get a registry size
static RegGetSize($reg)
{
  if ($reg < NN_al)
    return ($Bitness == 1) ? $Registers[$reg*RegObjectSize+RegSize]/2 : $Registers[$reg*RegObjectSize+RegSize];
  else
    return $Registers[$reg*RegObjectSize+RegSize];
}

//________________________________________________________________________________________________________________


//Taint the registry
static RegTaint($reg, $k)
{
  $TaintRegArray[$Registers[$reg*RegObjectSize+RegOffset] + $k] = Tainted;
  return Tainted;
}

//________________________________________________________________________________________________________________


//Untaint the registry
static RegUntaint($reg, $k)
{
  $TaintRegArray[$Registers[$reg*RegObjectSize+RegOffset] + $k] = Untainted;
  return Untainted;
}

//________________________________________________________________________________________________________________


//Check for a tainted register
static RegCheck($reg, $k)
{
  return $TaintRegArray[$Registers[$reg*RegObjectSize+RegOffset] + $k];
}

//________________________________________________________________________________________________________________


//Taint the temporary registry
static RegTaintTMP($value, $dtyp)
{
  auto $k;
  for ($k=0; $k<(1 << ($dtyp & 3)); $k++)
    if (($value >> $k*8) & 0xff)
      RegTaint(NN_TMP, $k);
    else
      RegUntaint(NN_TMP, $k);
  return;
}

//________________________________________________________________________________________________________________


//Stub for the tainted checks
static StubCheck($addr, $k)
{
  return ($addr & TaintedAddress) ? Tainted : Untainted;
}

//________________________________________________________________________________________________________________


static Mov($cmd)
{
  //Define variables
  auto $k, $dest, $src, $fc2, $ft, $fu;
  auto $taint = Untainted;

  //Get the 1st operand
  if ($cmd.Op0.type == o_reg)
  {
    $dest = $cmd.Op0.reg; if ($dest >= NN_TMP) return Unsupported;
    $ft = RegTaint; $fu = RegUntaint;
  }
  else
  {
    $dest = GetPointerAddress($cmd.Op0, true);
    $ft = MemTaint; $fu = MemUntaint;
  }

  //Get the 2nd operand
  if ($cmd.Op1.type == o_imm)
    $fc2 = StubCheck;
  else if ($cmd.Op1.type == o_reg)
  {
    $src = $cmd.Op1.reg; if ($src >= NN_TMP) return Unsupported;
    $fc2 = RegCheck;
  }
  else
  {
    $src = GetPointerAddress($cmd.Op1, true);
    $fc2 = MemCheck;
  }

  //Process the size of operands: dt_byte, dt_word, dt_dword, dt_qword
  for ($k=0; $k<(1 << ($cmd.Op0.dtyp & 3)); $k++)
    if ($fc2($src, $k))
      $taint = $taint | $ft($dest, $k);
    else
      $taint = $taint | $fu($dest, $k);

  //Untaint the high half of 64-bits register
  if (($cmd.Op0.type == o_reg) && ($cmd.Op0.dtyp == dt_dword))
    for ($k=4; $k<8; $k++)
      $fu($dest, $k);

  //Exit
  return $taint;
}

//________________________________________________________________________________________________________________


//Instructions: 386
static Instructions386($ea, $cmd)
{
  //Define variables
  auto $i, $j, $k, $dest, $src, $size, $fc1, $fc2, $ft, $fu, $fl;
  auto $taint = Untainted;

  //j*
  if ((($cmd.itype >= NN_ja) && ($cmd.itype <= NN_jc)) || (($cmd.itype >= NN_je) && ($cmd.itype <= NN_jz))) return Untainted;
  if (($cmd.itype >= NN_jmp) && ($cmd.itype <= NN_jmpshort)) return Untainted;

  //mov
  if (($cmd.itype == NN_mov) || (($cmd.itype >= NN_lgs) && ($cmd.itype <= NN_lfs)))
    return Mov($cmd);

  //push
  if ($cmd.itype == NN_push)
  {
    //Get the address of the stack
    $dest = GetTevRegVal(TEV_NONE, RegGetName(NN_sp)) - (1 << ($cmd.Op0.dtyp & 3));
    //Get an operand type
    if ($cmd.Op0.type == o_imm)
      $fc1 = StubCheck;
    else if ($cmd.Op0.type == o_reg)
    {
      $src = $cmd.Op0.reg; if ($src >= NN_TMP) return Unsupported;
      $fc1 = RegCheck;
    }
    else
    {
      $src = GetPointerAddress($cmd.Op0, true);
      $fc1 = MemCheck;
    }
    //Test for tainted: NN_sp
    for ($k=0; $k<(1 << ($cmd.Op0.dtyp & 3)); $k++)
      $fl = $fl | RegCheck(NN_sp, $k);
    //Process the size of address
    for ($k=0; $k<(1 << ($cmd.Op0.dtyp & 3)); $k++)
      if ($fc1($src, $k) || $fl)
        $taint = $taint | MemTaint($dest, $k);
      else
        $taint = $taint | MemUntaint($dest, $k);
    //Exit
    return $taint;
  }

  //pop
  if ($cmd.itype == NN_pop)
  {
    //Get the address of the stack
    $src = GetTevRegVal(TEV_NONE, RegGetName(NN_sp));
    //Get the operand of a command
    if ($cmd.Op0.type == o_reg)
    {
      $dest = $cmd.Op0.reg; if ($dest >= NN_TMP) return Unsupported;
      $ft = RegTaint; $fu = RegUntaint;
    }
    else
    {
      $dest = GetPointerAddress($cmd.Op0, true);
      $ft = MemTaint; $fu = MemUntaint;
    }
    //Test for tainted: NN_sp
    for ($k=0; $k<(1 << ($cmd.Op0.dtyp & 3)); $k++)
      $fl = $fl | RegCheck(NN_sp, $k);
    //Process the size of address
    for ($k=0; $k<(1 << ($cmd.Op0.dtyp & 3)); $k++)
      if (MemCheck($src, $k) || $fl)
        $taint = $taint | $ft($dest, $k);
      else
        $taint = $taint | $fu($dest, $k);
    //Exit
    return $taint;
  }

  //call
  if (($cmd.itype >= NN_call) && ($cmd.itype <= NN_callni))
  {
    //Get the address of the stack
    $dest = GetTevRegVal(TEV_NONE, RegGetName(NN_sp)) - (GetSegmentAttr($ea, SEGATTR_BITNESS) << 2);
    //Test for the tainted calling address
    if ($cmd.Op0.type == o_near)
    {
      $src = 0;
      $fc1 = StubCheck;
    }
    else if ($cmd.Op0.type == o_reg)
    {
      $src = $cmd.Op0.reg;
      $fc1 = RegCheck;
      if ($Segment != SegName(GetTevRegVal(TEV_NONE, RegGetName($cmd.Op0.reg))))
      {
        $TracingStop = true;
	$CalleeFunc = GetTevRegVal(TEV_NONE, RegGetName($cmd.Op0.reg));
      }
    }
    else
    {
      $src = GetPointerAddress($cmd.Op0, true);
      $fc1 = MemCheck;
      if ($Segment != SegName($cmd.Op0.addr))
      {
        $TracingStop = true;
	$CalleeFunc = $cmd.Op0.addr;
      }
    }
    //Untaint the stack
    for ($k=0; $k<(GetSegmentAttr($ea, SEGATTR_BITNESS) << 2); $k++)
      $taint = $taint | MemUntaint($dest, $k) | $fc1($src, $k) | RegCheck(NN_sp, $k);

    //Exit
    return $taint;
  }

  //ret
  if ($cmd.itype == NN_retn)
  {
    //Get the address of the stack
    $src = GetTevRegVal(TEV_NONE, RegGetName(NN_sp));
    //Test for the tainted return address
    for ($k=0; $k<(GetSegmentAttr($ea, SEGATTR_BITNESS) << 2); $k++)
      $taint = $taint | MemCheck($src, $k) | RegCheck(NN_sp, $k);
    //Exit
    return $taint;
  }

  //neg, not, inc, dec
  if (($cmd.itype == NN_neg) || ($cmd.itype == NN_not) || ($cmd.itype == NN_inc) || ($cmd.itype == NN_dec))
  {
    //Get the operand of a command
    if ($cmd.Op0.type == o_reg)
    {
      $dest = $cmd.Op0.reg;
      $fc1 = RegCheck;
    }
    else
    {
      $dest = GetPointerAddress($cmd.Op0, true);
      $fc1 = MemCheck;
    }
    //Process the size of an operand
    for ($k=0; $k<(1 << ($cmd.Op0.dtyp & 3)); $k++)
      $taint = $taint | $fc1($dest, $k);
    //Exit
    return $taint;
  }

  //and
  if ($cmd.itype == NN_and)
  {
    //Get the 1st operand
    if ($cmd.Op0.type == o_reg)
    {
      $dest = $cmd.Op0.reg;
      $ft = RegTaint; $fu = RegUntaint;
      $fc1 = RegCheck;
    }
    else
    {
      $dest = GetPointerAddress($cmd.Op0, true);
      $ft = MemTaint; $fu = MemUntaint;
      $fc1 = MemCheck;
    }
    //Get the 2nd operand
    if ($cmd.Op1.type == o_imm)
    {
      //Generate a value in the temporary register
      RegTaintTMP($cmd.Op1.value, $cmd.Op1.dtyp);
      $src = NN_TMP;
      $fc2 = RegCheck;
    }
    else if ($cmd.Op1.type == o_reg)
    {
      //Generate a value in the temporary register
      RegTaintTMP(GetTevRegVal(TEV_NONE, RegGetName($cmd.Op1.reg)), $cmd.Op1.dtyp);
      $src = NN_TMP;
      $fc2 = RegCheck;
    }
    else
    {
      $src = GetPointerAddress($cmd.Op1, true);
      $fc2 = MemCheck;
    }
    //Process the size of operands
    for ($k=0; $k<(1 << ($cmd.Op0.dtyp & 3)); $k++)
      if ($fc1($dest, $k) && $fc2($src, $k))
        $taint = $taint | $ft($dest, $k);
      else
        $taint = $taint | $fu($dest, $k);
    //Untaint the high half of 64-bits register
    if (($cmd.Op0.type == o_reg) && ($cmd.Op0.dtyp == dt_dword))
      for ($k=4; $k<8; $k++)
        $fu($dest, $k);
    //Exit
    return $taint;
  }

  //adc, add, or, sbb, sub
  if (($cmd.itype == NN_add) || ($cmd.itype == NN_adc) || ($cmd.itype == NN_or) || ($cmd.itype == NN_sbb) || ($cmd.itype == NN_sub))
  {
    //Get the 1st operand
    if ($cmd.Op0.type == o_reg)
    {
      $dest = $cmd.Op0.reg;
      $ft = RegTaint; $fu = RegUntaint;
      $fc1 = RegCheck;
    }
    else
    {
      $dest = GetPointerAddress($cmd.Op0, true);
      $ft = MemTaint; $fu = MemUntaint;
      $fc1 = MemCheck;
    }
    //Get the 2nd operand
    if ($cmd.Op1.type == o_imm)
      $fc2 = StubCheck;
    else if ($cmd.Op1.type == o_reg)
    {
      $src = $cmd.Op1.reg;
      $fc2 = RegCheck;
    }
    else
    {
      $src = GetPointerAddress($cmd.Op1, true);
      $fc2 = MemCheck;
    }
    //Process the size of operands
    for ($k=0; $k<(1 << ($cmd.Op0.dtyp & 3)); $k++)
      if ($fc1($dest, $k) || $fc2($src, $k))
      {
        if (($dest == $src) && (($cmd.itype == NN_sbb) || ($cmd.itype == NN_sub)))
          $taint = $taint | $fu($dest, $k);
	else
          $taint = $taint | $ft($dest, $k);
      }
    //Untaint the high half of 64-bits register
    if (($cmd.Op0.type == o_reg) && ($cmd.Op0.dtyp == dt_dword))
      for ($k=4; $k<8; $k++)
        $fu($dest, $k);
    //Exit
    return $taint;
  }

  //xor
  if ($cmd.itype == NN_xor)
  {
    //Get the 1st operand
    if ($cmd.Op0.type == o_reg)
    {
      $dest = $cmd.Op0.reg;
      $ft = RegTaint; $fu = RegUntaint;
      $fc1 = RegCheck;
    }
    else
    {
      $dest = GetPointerAddress($cmd.Op0, true);
      $ft = MemTaint; $fu = MemUntaint;
      $fc1 = MemCheck;
    }
    //Get the 2nd operand
    if ($cmd.Op1.type == o_imm)
      $fc2 = StubCheck;
    else if ($cmd.Op1.type == o_reg)
    {
      $src = $cmd.Op1.reg;
      $fc2 = RegCheck;
    }
    else
    {
      $src = GetPointerAddress($cmd.Op1, true);
      $fc2 = MemCheck;
    }
    //Process the size of operands
    for ($k=0; $k<(1 << ($cmd.Op0.dtyp & 3)); $k++)
      if (($fc1($dest, $k) || $fc2($src, $k)) && ($dest != $src))
        $taint = $taint | $ft($dest, $k);
      else
        $taint = $taint | $fu($dest, $k);
    //Untaint the high half of 64-bits register
    if (($cmd.Op0.type == o_reg) && ($cmd.Op0.dtyp == dt_dword))
      for ($k=4; $k<8; $k++)
        $fu($dest, $k);
    //Exit
    return $taint;
  }

  //cmp, test, bsf, bsr, bt, btc, btr, bts
  if (($cmd.itype == NN_cmp) || ($cmd.itype == NN_test) || (($cmd.itype >= NN_bsf) && ($cmd.itype <= NN_bts)))
  {
    //Get the 1st operand
    if ($cmd.Op0.type == o_reg)
    {
      $dest = $cmd.Op0.reg;
      $fc1 = RegCheck;
    }
    else
    {
      $dest = GetPointerAddress($cmd.Op0, true);
      $fc1 = MemCheck;
    }
    //Get the 2nd operand
    if ($cmd.Op1.type == o_imm)
      $fc2 = StubCheck;
    else if ($cmd.Op1.type == o_reg)
    {
      $src = $cmd.Op1.reg;
      $fc2 = RegCheck;
    }
    else
    {
      $src = GetPointerAddress($cmd.Op1, true);
      $fc2 = MemCheck;
    }
    //Process the size of operands
    for ($k=0; $k<(1 << ($cmd.Op0.dtyp & 3)); $k++)
      $taint = $taint | $fc1($dest, $k) | $fc2($src, $k);
    //Exit
    return $taint;
  }

  //lea
  if ($cmd.itype == NN_lea)
  {
    //Get the 1st operand
    $dest = $cmd.Op0.reg;
    $ft = RegTaint; $fu = RegUntaint;
    //Get the 2nd operand
    if ($cmd.Op1.type == o_reg)
    {
      $src = $cmd.Op1.reg;
      $fc2 = RegCheck;
    }
    else
    {
      $src = GetPointerAddress($cmd.Op1, false);
      $fc2 = StubCheck;
    }
    //Process the size of operands
    for ($k=0; $k<(1 << ($cmd.Op0.dtyp & 3)); $k++)
      if ($fc2($src, $k))
        $taint = $taint | $ft($dest, $k);
      else
        $taint = $taint | $fu($dest, $k);
    //Untaint the high half of 64-bits register
    if (($cmd.Op0.type == o_reg) && ($cmd.Op0.dtyp == dt_dword))
      for ($k=4; $k<8; $k++)
        $fu($dest, $k);
    //Exit
    return $taint;
  }

  //movs
  if ($cmd.itype == NN_movs)
  {
    //Initialize values
    $dest = $cmd.Op0.reg;
    $src = $cmd.Op1.reg;
    $fl = GetTevRegVal(TEV_NONE, "DF") ? -1 : 1;
    $size = 1 << ($cmd.Op1.dtyp & 3);
    //Check operands
    if ((Byte($cmd.ea) == NN_repe) || (Byte($cmd.ea) == NN_repne))
    {
      $j = GetTevRegVal(TEV_NONE, RegGetName(NN_cx));
      for ($k=0; $k<RegGetSize($dest); $k++)
	$taint = $taint | RegCheck($dest, $k) | RegCheck($src, $k) | RegCheck(NN_cx, $k) | RegUntaint(NN_cx, $k);
    }
    else
    {
      $j = 1;
      for ($k=0; $k<RegGetSize($dest); $k++)
	$taint = $taint | RegCheck($dest, $k) | RegCheck($src, $k);
    }
    //Taint memory
    if ($taint)
    {
      for ($i=0; $i<$j; $i++)
	for ($k=0; $k<$size; $k++)
          MemTaint(GetTevRegVal(TEV_NONE, RegGetName($dest)) + $fl*$i*$size, $k);
    }
    else
    {
      for ($i=0; $i<$j; $i++)
	for ($k=0; $k<$size; $k++)
	  if (MemCheck(GetTevRegVal(TEV_NONE, RegGetName($src)) + $fl*$i*$size, $k))
	    $taint = $taint | MemTaint(GetTevRegVal(TEV_NONE, RegGetName($dest)) + $fl*$i*$size, $k);
	  else
	    MemUntaint(GetTevRegVal(TEV_NONE, RegGetName($dest)) + $fl*$i*$size, $k);
    }
    //Exit
    return $taint;
  }

  //stos
  if ($cmd.itype == NN_stos)
  {
    //Initialize values
    $dest = $cmd.Op0.reg;
    $fl = GetTevRegVal(TEV_NONE, "DF") ? -1 : 1;
    $size = 1 << ($cmd.Op1.dtyp & 3);
    //Check operands
    if ((Byte($cmd.ea) == NN_repe) || (Byte($cmd.ea) == NN_repne))
    {
      $j = GetTevRegVal(TEV_NONE, RegGetName(NN_cx));
      for ($k=0; $k<RegGetSize($dest); $k++)
	$taint = $taint | RegCheck($dest, $k) | RegCheck(NN_cx, $k) | RegUntaint(NN_cx, $k);
    }
    else
    {
      $j = 1;
      for ($k=0; $k<RegGetSize($dest); $k++)
	$taint = $taint | RegCheck($dest, $k);
    }
    //Taint memory
    if ($taint)
    {
      for ($i=0; $i<$j; $i++)
	for ($k=0; $k<$size; $k++)
          MemTaint(GetTevRegVal(TEV_NONE, RegGetName($dest)) + $fl*$i*$size, $k);
    }
    else
    {
      for ($i=0; $i<$j; $i++)
	for ($k=0; $k<$size; $k++)
	  if (RegCheck(NN_ax, $k))
	    $taint = $taint | MemTaint(GetTevRegVal(TEV_NONE, RegGetName($dest)) + $fl*$i*$size, $k);
	  else
	    MemUntaint(GetTevRegVal(TEV_NONE, RegGetName($dest)) + $fl*$i*$size, $k);
    }
    //Exit
    return $taint;
  }

  //lods
  if ($cmd.itype == NN_lods)
  {
    //Initialize values
    $src = $cmd.Op1.reg;
    $fl = GetTevRegVal(TEV_NONE, "DF") ? -1 : 1;
    $size = 1 << ($cmd.Op0.dtyp & 3);
    //Check operands
    if ((Byte($cmd.ea) == NN_repe) || (Byte($cmd.ea) == NN_repne))
    {
      $j = GetTevRegVal(TEV_NONE, RegGetName(NN_cx));
      for ($k=0; $k<RegGetSize($src); $k++)
	$taint = $taint | RegCheck($src, $k) | RegCheck(NN_cx, $k) | RegUntaint(NN_cx, $k);
    }
    else
    {
      $j = 1;
      for ($k=0; $k<RegGetSize($src); $k++)
	$taint = $taint | RegCheck($src, $k);
    }
    //Taint the register: NN_ax
    if ($taint)
      for ($k=0; $k<$size; $k++) RegTaint(NN_ax, $k);
    else if ($j != 0)
    {
      //Depends on the memory
      for ($k=0; $k<$size; $k++)
	if (MemCheck(GetTevRegVal(TEV_NONE, RegGetName($src)) + $fl*($j-1)*$size, $k))
	  $taint = $taint | RegTaint(NN_ax, $k);
	else
	  RegUntaint(NN_ax, $k);
    }
    //Untaint the high half of 64-bits register
    if (($j != 0) && ($cmd.Op0.dtyp == dt_dword))
      for ($k=4; $k<8; $k++)
        RegUntaint(NN_ax, $k);
    //Exit
    return $taint;
  }

  //cmps
  if ($cmd.itype == NN_cmps)
  {
    //Initialize values
    $dest = $cmd.Op1.reg;
    $src = $cmd.Op0.reg;
    $fl = GetTevRegVal(TEV_NONE, "DF") ? -1 : 1;
    $size = 1 << ($cmd.Op0.dtyp & 3);
    //Check operands
    if ((Byte($cmd.ea) == NN_repe) || (Byte($cmd.ea) == NN_repne))
    {
      $j = GetTevRegVal(TEV_NONE, RegGetName(NN_cx));
      for ($k=0; $k<RegGetSize($dest); $k++)
	$taint = $taint | RegCheck($dest, $k) | RegCheck($src, $k) | RegCheck(NN_cx, $k);
    }
    else
    {
      $j = 1;
      for ($k=0; $k<RegGetSize($dest); $k++)
	$taint = $taint | RegCheck($dest, $k) | RegCheck($src, $k);
    }
    //Check memory
    if (!$taint)
      for ($i=0; $i<$j; $i++)
	for ($k=0; $k<$size; $k++)
	  $taint = $taint | MemCheck(GetTevRegVal(TEV_NONE, RegGetName($dest)) + $fl*$i*$size, $k) | MemCheck(GetTevRegVal(TEV_NONE, RegGetName($src)) + $fl*$i*$size, $k);
    //Exit
    return $taint;
  }

  //scas
  if ($cmd.itype == NN_scas)
  {
    //Initialize values
    $dest = $cmd.Op0.reg;
    $fl = GetTevRegVal(TEV_NONE, "DF") ? -1 : 1;
    $size = 1 << ($cmd.Op1.dtyp & 3);
    //Check operands
    if ((Byte($cmd.ea) == NN_repe) || (Byte($cmd.ea) == NN_repne))
    {
      $j = GetTevRegVal(TEV_NONE, RegGetName(NN_cx));
      for ($k=0; $k<RegGetSize($dest); $k++)
	$taint = $taint | RegCheck($dest, $k) | RegCheck(NN_cx, $k);
      for ($k=0; $k<$size; $k++)
	$taint = $taint | RegCheck(NN_ax, $k);
    }
    else
    {
      $j = 1;
      for ($k=0; $k<RegGetSize($dest); $k++)
        $taint = $taint | RegCheck($dest, $k);
      for ($k=0; $k<$size; $k++)
        $taint = $taint | RegCheck(NN_ax, $k);
    }
    //Check memory
    if (!$taint)
      for ($i=0; $i<$j; $i++)
	for ($k=0; $k<$size; $k++)
	  $taint = $taint | MemCheck(GetTevRegVal(TEV_NONE, RegGetName($dest)) + $fl*$i*$size, $k);
    //Exit
    return $taint;
  }

  //rcl, rcr, rol, ror, sal, sar, shl, shr
  if ((($cmd.itype >= NN_rcl) && ($cmd.itype <= NN_ror)) || (($cmd.itype >= NN_sal) && ($cmd.itype <= NN_shr)))
  {
    //Get the operand
    if ($cmd.Op0.type == o_reg)
    {
      $dest = $cmd.Op0.reg;
      $ft = RegTaint;
      $fc1 = RegCheck;
    }
    else
    {
      $dest = GetPointerAddress($cmd.Op0, true);
      $ft = MemTaint;
      $fc1 = MemCheck;
    }
    //Process the counter
    if (($cmd.Op1.type == o_reg) && RegCheck(NN_cl, 0) && (GetTevRegVal(TEV_NONE, RegGetName(NN_cx)) & 0xff))
      $taint = Tainted;
    //Process the size of an operand
    for ($k=0; $k<(1 << ($cmd.Op0.dtyp & 3)); $k++)
      $taint = $taint | $fc1($dest, $k);
    if ($taint)
      for ($i=0; $i<(1 << ($cmd.Op0.dtyp & 3)); $i++)
        $ft($dest, $i);
    //Untaint the high half of 64-bits register
    if (($cmd.Op0.type == o_reg) && ($cmd.Op0.dtyp == dt_dword))
      for ($k=4; $k<8; $k++)
        RegUntaint($dest, $k);
    //Exit
    return $taint;
  }

  //imul
  if ($cmd.itype == NN_imul)
  {
    //Get operands
    $dest = $cmd.Op0.reg;
    $fc1 = RegCheck;
    if ($cmd.Op1.type == o_reg)
    {
      $src = $cmd.Op1.reg;
      $fc2 = RegCheck;
    }
    else
    {
      $src = GetPointerAddress($cmd.Op1, true);
      $fc2 = MemCheck;
    }
    //Process the number of operands
    $size = 1 << ($cmd.Op0.dtyp & 3);
    $k = ((Byte($cmd.ea) == 0x48) || (Byte($cmd.ea) == 0x66)) ? 1 : 0;
    if (($cmd.n == 2) && ((Byte($cmd.ea+$k) == 0xf6) || (Byte($cmd.ea+$k) == 0xf7)))
    {
      //One operand
      if ($size == 1)
      {
	if ($fc2($src, 0) | RegCheck(NN_al, 0))
	  $taint = $taint | RegTaint(NN_al, 0) | RegTaint(NN_ah, 0);
      }
      else
      {
	for ($k=0; $k<$size; $k++)
	  if ($fc2($src, $k) | RegCheck(NN_ax, $k))
	    for ($j=0; $j<$size; $j++)
	      $taint = $taint | RegTaint(NN_ax, $j) | RegTaint(NN_dx, $j);
      }
    }
    else if ($cmd.n == 2)
    {
      //Two operands
      for ($k=0; $k<$size; $k++)
        $taint = $taint | $fc1($dest, $k) | $fc2($src, $k);
      if ($taint)
	for ($j=0; $j<$size; $j++)
	  $taint = $taint | RegTaint($dest, $j);
    }
    else
    {
      //Three operands
      for ($k=0; $k<$size; $k++)
        $taint = $taint | $fc2($src, $k);
      for ($k=0; $k<$size; $k++)
        if ($taint)
          RegTaint($dest, $k);
	else
	  RegUntaint($dest, $k);
    }
    //Untaint the high half of 64-bits registers
    if (($cmd.Op0.type == o_reg) && ($cmd.Op0.dtyp == dt_dword))
      for ($k=4; $k<8; $k++)
      {
        RegUntaint($dest, $k);
	if (Byte($cmd.ea) == 0xf7) RegUntaint(NN_dx, $k);
      }
    //Exit
    return $taint;
  }

  //mul
  if ($cmd.itype == NN_mul)
  {
    //Get the operand
    if ($cmd.Op1.type == o_reg)
    {
      $src = $cmd.Op1.reg;
      $fc1 = RegCheck;
    }
    else
    {
      $src = GetPointerAddress($cmd.Op1, true);
      $fc1 = MemCheck;
    }
    //Process the size of operands
    $size = 1 << ($cmd.Op1.dtyp & 3);
    if ($size == 1)
    {
      if ($fc1($src, 0) | RegCheck(NN_al, 0))
        $taint = $taint | RegTaint(NN_al, 0) | RegTaint(NN_ah, 0);
    }
    else
    {
      for ($k=0; $k<$size; $k++)
        if ($fc1($src, $k) | RegCheck(NN_ax, $k))
          for ($j=0; $j<$size; $j++)
            $taint = $taint | RegTaint(NN_ax, $j) | RegTaint(NN_dx, $j);
    }
    //Untaint the high half of 64-bits registers
    if ($cmd.Op1.dtyp == dt_dword)
      for ($k=4; $k<8; $k++)
      {
        RegUntaint(NN_ax, $k);
	RegUntaint(NN_dx, $k);
      }
    //Exit
    return $taint;
  }

  //div, idiv
  if (($cmd.itype == NN_div) || ($cmd.itype == NN_idiv))
  {
    //Get the operand
    if ($cmd.Op1.type == o_reg)
    {
      $src = $cmd.Op1.reg;
      $fc1 = RegCheck;
    }
    else
    {
      $src = GetPointerAddress($cmd.Op1, true);
      $fc1 = MemCheck;
    }
    //Process the size of operands
    $size = 1 << ($cmd.Op1.dtyp & 3);
    if ($size == 1)
    {
      if ($fc1($src, 0) | RegCheck(NN_al, 0) | RegCheck(NN_ah, 0))
        $taint = $taint | RegTaint(NN_al, 0) | RegTaint(NN_ah, 0);
    }
    else
    {
      for ($k=0; $k<$size; $k++)
        if ($fc1($src, $k) | RegCheck(NN_ax, $k) | RegCheck(NN_dx, $k))
          for ($j=0; $j<$size; $j++)
            $taint = $taint | RegTaint(NN_ax, $j) | RegTaint(NN_dx, $j);
    }
    //Untaint the high half of 64-bits registers
    if ($cmd.Op1.dtyp == dt_dword)
      for ($k=4; $k<8; $k++)
      {
        RegUntaint(NN_ax, $k);
	RegUntaint(NN_dx, $k);
      }
    //Exit
    return $taint;
  }

  //movzx, movsx
  if (($cmd.itype == NN_movsx) || ($cmd.itype == NN_movzx))
  {
    //Get the 1st operand
    $dest = $cmd.Op0.reg;
    $ft = RegTaint; $fu = RegUntaint;
    //Get the 2nd operand
    if ($cmd.Op1.type == o_reg)
    {
      $src = $cmd.Op1.reg;
      $fc2 = RegCheck;
    }
    else
    {
      $src = GetPointerAddress($cmd.Op1, true);
      $fc2 = MemCheck;
    }
    //Process the size of operands
    for ($k=0; $k<(1 << ($cmd.Op0.dtyp & 3)); $k++)
      if ($fc2($src, $k) && ($k < (1 << ($cmd.Op1.dtyp & 3))))
	$taint = $taint | $ft($dest, $k);
      else
	$taint = $taint | $fu($dest, $k);
    //Untaint the high half of 64-bits register
    if ($cmd.Op0.dtyp == dt_dword)
      for ($k=4; $k<8; $k++)
        $fu($dest, $k);
    //Exit
    return $taint;
  }

  //xchg
  if ($cmd.itype == NN_xchg)
  {
    //Get the 1st operand
    if ($cmd.Op0.type == o_reg)
    {
      $dest = $cmd.Op0.reg;
      $fc1 = RegCheck;
      $ft = RegTaint; $fu = RegUntaint;
    }
    else
    {
      $dest = GetPointerAddress($cmd.Op0, true);
      $fc1 = MemCheck;
      $ft = MemTaint; $fu = MemUntaint;
    }
    //Get the 2nd operand
    if ($cmd.Op1.type == o_reg)
    {
      $src = $cmd.Op1.reg;
      $fc2 = RegCheck;
      $i = RegTaint; $j = RegUntaint;
    }
    else
    {
      $src = GetPointerAddress($cmd.Op1, true);
      $fc2 = MemCheck;
      $i = MemTaint; $j = MemUntaint;
    }
    //Process the size of operands
    for ($k=0; $k<(1 << ($cmd.Op0.dtyp & 3)); $k++)
    {
      //Read the 1st operand before the changing
      if ($fc1($dest, $k))
        RegTaint(NN_TMP, $k);
      else
        RegUntaint(NN_TMP, $k);
      //Change the 1st operand
      if ($fc2($src, $k))
        $taint = $taint | $ft($dest, $k);
      else
        $taint = $taint | $fu($dest, $k);
      //Change the 2nd operand
      if (RegCheck(NN_TMP, $k))
        $taint = $taint | $i($src, $k);
      else
        $taint = $taint | $j($src, $k);
    }
    //Untaint the high half of 64-bits registers
    for ($k=4; $k<8; $k++)
    {
      if (($cmd.Op0.type == o_reg) && ($cmd.Op0.dtyp == dt_dword))
        $fu($dest, $k);
      if (($cmd.Op1.type == o_reg) && ($cmd.Op1.dtyp == dt_dword))
        $j($src, $k);
    }
    //Exit
    return $taint;
  }

  //loopXX
  if (($cmd.itype >= NN_loopw) && ($cmd.itype <= NN_loopdne))
  {
    //Define the size of the instruction
    if (($cmd.itype == NN_loopw) || ($cmd.itype == NN_loopwe) || ($cmd.itype == NN_loopwne))
      $j = 1 << (dt_word & 3);
    else if (($cmd.itype == NN_loopd) || ($cmd.itype == NN_loopde) || ($cmd.itype == NN_loopdne))
      $j = 1 << (dt_dword & 3);
    else if (($cmd.itype == NN_loopq) || ($cmd.itype == NN_loopqe) || ($cmd.itype == NN_loopqne))
      $j = 1 << (dt_qword & 3);
    else
      $j = 1 << ($cmd.Op1.dtyp & 3);
    $i = 0;
    for ($k=0; $k<$j; $k++)
      $i = $i << 8 | 0xff;
    //Process the size of register
    for ($k=0; $k<$j; $k++)
      $taint = $taint | RegCheck(NN_cx, $k);
    //Untaint the register for the last iteration
    if ($taint && ((GetTevRegVal(TEV_NONE, RegGetName(NN_cx)) & $i) == 1))
      for ($k=0; $k<$j; $k++)
        RegUntaint(NN_cx, $k);
    //Exit
    return $taint;
  }

  //jXcxz
  if (($cmd.itype >= NN_jcxz) && ($cmd.itype <= NN_jrcxz))
  {
    if ($cmd.itype == NN_jcxz)
      $j = 1 << (dt_word & 3);
    else if ($cmd.itype == NN_jecxz)
      $j = 1 << (dt_dword & 3);
    else
      $j = 1 << (dt_qword & 3);
    //Process the size of register
    for ($k=0; $k<$j; $k++)
      $taint = $taint | RegCheck(NN_cx, $k);
    //Exit
    return $taint;
  }

  //cbw, cwde, cdqe
  if (($cmd.itype >= NN_cbw) && ($cmd.itype <= NN_cdqe))
  {
    //Process the size of register
    $size = 1 << ($cmd.Op0.dtyp & 3);
    for ($k=0; $k<($size*2); $k++)
      if ($k < $size)
        $taint = $taint | RegCheck(NN_ax, $k);
      else
        RegUntaint(NN_ax, $k);
    //Untaint the high half of 64-bits register
    if ($cmd.Op0.dtyp == dt_word)
      for ($k=4; $k<8; $k++)
        RegUntaint(NN_ax, $k);
    //Exit
    return $taint;
  }

  //cwd, cdq, cqo
  if (($cmd.itype >= NN_cwd) && ($cmd.itype <= NN_cqo))
  {
    //Process the size of register
    for ($k=0; $k<(1 << ($cmd.Op0.dtyp & 3)); $k++)
      $taint = $taint | RegCheck(NN_ax, $k) | RegUntaint(NN_dx, $k);
    //Untaint the high half of 64-bits register
    if ($cmd.Op0.dtyp == dt_dword)
      for ($k=4; $k<8; $k++)
        RegUntaint(NN_dx, $k);
    //Exit
    return $taint;
  }

  //pusha
  if (($cmd.itype >= NN_pushaw) && ($cmd.itype <= NN_pushad))
  {
    //Get the size of registers
    $size = ($cmd.itype == NN_pushaw) ? 2 : 4;
    //Process the memory for the stack
    for ($i=0; $i<8; $i++)
    {
      //Get the address of the stack
      $dest = GetTevRegVal(TEV_NONE, RegGetName(NN_sp)) - ($i+1)*$size;
      //Process the size of address
      for ($k=0; $k<$size; $k++)
	if (RegCheck($i, $k))
	  $taint = $taint | MemTaint($dest, $k);
	else
	  $taint = $taint | MemUntaint($dest, $k);
    }
    //Exit
    return $taint;
  }

  //popa
  if (($cmd.itype >= NN_popaw) && ($cmd.itype <= NN_popad))
  {
    //Get the size of registers
    $size = ($cmd.itype == NN_popaw) ? 2 : 4;
    //Process the memory for the stack
    for ($i=8; $i>0; $i--)
    {
      //Get the address of the stack
      $dest = GetTevRegVal(TEV_NONE, RegGetName(NN_sp)) + (8-$i)*$size;
      //Process the size of address
      for ($k=0; $k<$size; $k++)
	if (MemCheck($dest, $k))
	  $taint = $taint | RegTaint($i-1, $k);
	else
	  RegUntaint($i-1, $k);
    }
    //Exit
    return $taint;
  }

  //NN_seta, ...
  if (($cmd.itype >= NN_setae) && ($cmd.itype <= NN_setz))
  {
    //Exit
    return RegUntaint($cmd.Op0.reg, 0);
  }

  //nop, stc, std, clc, cld, cmc
  if (($cmd.itype == NN_nop) || ($cmd.itype == NN_stc) || ($cmd.itype == NN_std) || ($cmd.itype == NN_clc) || ($cmd.itype == NN_cld) || ($cmd.itype == NN_cmc)) return Untainted;

  //enter
  if (($cmd.itype >= NN_enterw) && ($cmd.itype <= NN_enterq))
  {
    //Get the size
    $size = ($cmd.itype == NN_enterw) ? 2 : (($cmd.itype == NN_enterd) ? 4 : 8);
    //Get the address of the stack
    $dest = GetTevRegVal(TEV_NONE, RegGetName(NN_sp)) - $size;
    //Process the size of an address
    for ($k=0; $k<$size; $k++)
      if (RegCheck(NN_bp, $k))
        $taint = $taint | MemTaint($dest, $k);
      else
        $taint = $taint | MemUntaint($dest, $k);
    //Process the size of operands
    for ($k=0; $k<$size; $k++)
      if (RegCheck(NN_sp, $k))
        $taint = $taint | RegTaint(NN_bp, $k);
      else
        $taint = $taint | RegUntaint(NN_bp, $k);
    //Exit
    return $taint;
  }

  //leave
  if (($cmd.itype >= NN_leavew) && ($cmd.itype <= NN_leaveq))
  {
    //Get the size
    $size = ($cmd.itype == NN_leavew) ? 2 : (($cmd.itype == NN_leaved) ? 4 : 8);
    //Process the size of operands
    for ($k=0; $k<$size; $k++)
      if (RegCheck(NN_bp, $k))
        $taint = $taint | RegTaint(NN_sp, $k);
      else
        $taint = $taint | RegUntaint(NN_sp, $k);
    //Get the address of the stack
    $src = GetTevRegVal(TEV_NONE, RegGetName(NN_sp));
    //Process the size of address
    for ($k=0; $k<$size; $k++)
      if (MemCheck($src, $k))
        $taint = $taint | RegTaint(NN_bp, $k);
      else
        $taint = $taint | RegUntaint(NN_bp, $k);
    //Exit
    return $taint;
  }

  //shld, shrd
  if (($cmd.itype == NN_shld) || ($cmd.itype == NN_shrd))
  {
    //Get the 1st operand
    if ($cmd.Op0.type == o_reg)
    {
      $dest = $cmd.Op0.reg;
      $ft = RegTaint;
      $fc1 = RegCheck;
    }
    else
    {
      $dest = GetPointerAddress($cmd.Op0, true);
      $ft = MemTaint;
      $fc1 = MemCheck;
    }
    //Get the 2nd operand
    $src = $cmd.Op1.reg;
    $fc2 = RegCheck;
    //Process the counter
    if (($cmd.Op2.type == o_reg) && RegCheck(NN_cl, 0) && (GetTevRegVal(TEV_NONE, RegGetName(NN_cx)) & 0xff))
      $taint = Tainted;
    //Test operands for tainted
    for ($k=0; $k<(1 << ($cmd.Op0.dtyp & 3)); $k++)
      $taint = $taint | $fc1($dest, $k) | $fc2($src, $k);
    if ($taint)
      for ($i=0; $i<(1 << ($cmd.Op0.dtyp & 3)); $i++)
        $ft($dest, $i);
    //Untaint the high half of 64-bits register
    if (($cmd.Op0.type == o_reg) && ($cmd.Op0.dtyp == dt_dword))
      for ($k=4; $k<8; $k++)
        RegUntaint($dest, $k);
    //Exit
    return $taint;
  }

  //pushfw, pushf, pushfd, pushfq
  if (($cmd.itype >= NN_pushfw) && ($cmd.itype <= NN_pushfq))
  {
    //Get the size
    $size = ($cmd.itype == NN_pushfw) ? 2 : (($cmd.itype == NN_pushfd) ? 4 : 8);
    //Process the size of address
    for ($k=0; $k<$size; $k++)
      MemUntaint((GetTevRegVal(TEV_NONE, RegGetName(NN_sp)) - $size), $k);
    //Exit
    return Untainted;
  }

  //popfw, popf, popfd, popfq
  if (($cmd.itype >= NN_popfw) && ($cmd.itype <= NN_popfq))
  {
    //Get the size
    $size = ($cmd.itype == NN_popfw) ? 2 : (($cmd.itype == NN_popfq) ? 4 : 8);
    //Process the size of address
    for ($k=0; $k<$size; $k++)
      if (MemCheck(GetTevRegVal(TEV_NONE, RegGetName(NN_sp)), $k))
        $taint = Tainted;
    //Exit
    return $taint;
  }

  //lahf
  if ($cmd.itype == NN_lahf)
  {
    //Exit
    return RegUntaint(NN_ah, 0);
  }

  //sahf
  if ($cmd.itype == NN_sahf)
  {
    //Exit
    return RegCheck(NN_ah, 0);
  }

  //aaa, aad, aam, aas, daa, das
  if ((($cmd.itype >= NN_aaa) && ($cmd.itype <= NN_aas)) || ($cmd.itype == NN_daa) || ($cmd.itype == NN_das))
  {
    //Exit
    return RegCheck(NN_al, 0);
  }

  //bound
  if ($cmd.itype == NN_bound)
  {
    //Process the size of the operand
    for ($k=0; $k<(1 << ($cmd.Op0.dtyp & 3)); $k++)
      $taint = $taint | RegCheck($cmd.Op0.reg, $k);
    //Exit
    return $taint;
  }

  //smsw
  if ($cmd.itype == NN_smsw)
  {
    //Get the operand of a command
    if ($cmd.Op0.type == o_reg)
    {
      $dest = $cmd.Op0.reg;
      $fu = RegUntaint;
    }
    else
    {
      $dest = GetPointerAddress($cmd.Op0, true);
      $fu = MemUntaint;
    }
    //Process the size of an operand
    for ($k=0; $k<(dt_word*2); $k++)
      $fu($dest, $k);
    //Exit
    return Untainted;
  }

  //Exit
  return Unsupported;
}

//________________________________________________________________________________________________________________


//Instructions: Pentium
static InstructionsPentium($ea, $cmd)
{
  //Define variables
  auto $k;

  //rdtsc
  if ($cmd.itype == NN_rdtsc)
  {
    for ($k=0; $k<RegGetSize(NN_ax); $k++)
    {
      RegUntaint(NN_ax, $k);
      RegUntaint(NN_dx, $k);
    }
    //Exit
    return Untainted;
  }

  //Exit
  return Unsupported;
}

//________________________________________________________________________________________________________________


//Instructions: Pentium Pro
static InstructionsPentiumPro($ea, $cmd)
{
  //Define variables
  auto $cf, $pf, $zf, $sf, $of;

  //Get flags
  $cf = GetTevRegVal(TEV_NONE, "CF");
  $pf = GetTevRegVal(TEV_NONE, "PF");
  $zf = GetTevRegVal(TEV_NONE, "ZF");
  $sf = GetTevRegVal(TEV_NONE, "SF");
  $of = GetTevRegVal(TEV_NONE, "OF");

  //cmova (cf=0 & zf=0)
  if (($cmd.itype == NN_cmova) && !$cf && !$zf)
    return Mov($cmd);

  //cmovb (cf=1)
  if (($cmd.itype == NN_cmovb) && $cf)
    return Mov($cmd);

  //cmovbe (cf=1 | zf=1)
  if (($cmd.itype == NN_cmovbe) && ($cf || $zf))
    return Mov($cmd);

  //cmovg (zf=0 & sf=of)
  if (($cmd.itype == NN_cmovg) && !$zf && ($sf == $of))
    return Mov($cmd);

  //cmovge (sf=of)
  if (($cmd.itype == NN_cmovge) && ($sf == $of))
    return Mov($cmd);

  //cmovl (sf!=of)
  if (($cmd.itype == NN_cmovl) && ($sf != $of))
    return Mov($cmd);

  //cmovle (zf=1 | sf!=of)
  if (($cmd.itype == NN_cmovle) && ($zf || ($sf != $of)))
    return Mov($cmd);

  //cmovnb (cf=0)
  if (($cmd.itype == NN_cmovnb) && !$cf)
    return Mov($cmd);

  //cmovno (of=0)
  if (($cmd.itype == NN_cmovno) && !$of)
    return Mov($cmd);

  //cmovnp (pf=0)
  if (($cmd.itype == NN_cmovnp) && !$pf)
    return Mov($cmd);

  //cmovns (sf=0)
  if (($cmd.itype == NN_cmovns) && !$sf)
    return Mov($cmd);

  //cmovnz (zf=0)
  if (($cmd.itype == NN_cmovnz) && !$zf)
    return Mov($cmd);

  //cmovo (of=1)
  if (($cmd.itype == NN_cmovo) && $of)
    return Mov($cmd);

  //cmovp (pf=1)
  if (($cmd.itype == NN_cmovp) && $pf)
    return Mov($cmd);

  //cmovs (sf=1)
  if (($cmd.itype == NN_cmovs) && $sf)
    return Mov($cmd);

  //cmovz (zf=1)
  if (($cmd.itype == NN_cmovz) && $zf)
    return Mov($cmd);

  //cmovX
  if (($cmd.itype >= NN_cmova) && ($cmd.itype <= NN_cmovz))
    return Untainted;

  //Exit
  return Unsupported;
}

//________________________________________________________________________________________________________________


//Action of an instruction
static InstructionAction($ea, $cmd)
{
  //Define variables
  auto $taint;

  //Instructions
  $taint = Instructions386($ea, $cmd);
  if ($taint == Unsupported)
    $taint = InstructionsPentium($ea, $cmd);
  if ($taint == Unsupported)
    $taint = InstructionsPentiumPro($ea, $cmd);

  //Exit
  return $taint;
}

//________________________________________________________________________________________________________________


//Taint registers by a user
static RegTaintUser()
{
  //Define variables
  auto $i, $k, $e;
  auto $Regs = "";

  //Taint registers
  try
  {
    Message("Get: initial tainted registers ...");
    if ($Bitness == 1)
      $e = AskStr("", "Registers (eax ecx edx ebx esp ebp esi edi al cl dl bl ah ch dh bh spl bpl sil dil)");
    else
      $e = AskStr("", "Registers (rax rcx rdx rbx rsp rbp rsi rdi r8 r9 r10 r11 r12 r13 r14 r15 al cl dl bl ah ch dh bh spl bpl sil dil)");
    if ($e == "") throw 0;
    for ($i=0; $i<NN_TMP; $i++)
      if (strstr($e, RegGetName($i)) != -1)
      {
	$Regs = $Regs + RegGetName($i) + " ";
	for ($k=0; $k<RegGetSize($i); $k++)
	  RegTaint($i, $k);
      }
    if ($Regs != "")
      Message("done\n" + "Tainted registers: " + $Regs + "\n");
    else
      throw 0;
  }
  catch($e)
  {
    Message("failed\n");
  }
}

//________________________________________________________________________________________________________________


//Untaint registers by a user
static RegUntaintUser()
{
  //Define variables
  auto $i, $k, $e;
  auto $Regs = "";

  //Taint registers
  try
  {
    Message("Get: initial untainted registers ...");
    if ($Bitness == 1)
      $e = AskStr("", "Registers (eax ecx edx ebx esp ebp esi edi al cl dl bl ah ch dh bh spl bpl sil dil)");
    else
      $e = AskStr("", "Registers (rax rcx rdx rbx rsp rbp rsi rdi r8 r9 r10 r11 r12 r13 r14 r15 al cl dl bl ah ch dh bh spl bpl sil dil)");
    if ($e == "") throw 0;
    for ($i=0; $i<NN_TMP; $i++)
      if (strstr($e, RegGetName($i)) != -1)
      {
	$Regs = $Regs + RegGetName($i) + " ";
	for ($k=0; $k<RegGetSize($i); $k++)
	  RegUntaint($i, $k);
      }
    if ($Regs != "")
      Message("done\n" + "Untainted registers: " + $Regs + "\n");
    else
      throw 0;
  }
  catch($e)
  {
    Message("failed\n");
  }
}

//________________________________________________________________________________________________________________


//Taint memory by a user
static MemTaintUser()
{
  //Define variables
  auto $i, $e, $dest, $size;

  //Taint a buffer
  try
  {
    //Get the memory pointer
    Message("Get: initial tainted memory ...");
    $dest = AskAddr("", "Pointer to memory");
    if (!($dest + 1)) throw 0;
    $size = AskLong("", "Memory size");
    if (!($size+1)) throw 0;

    //Taint the range
    for ($i=0; $i<$size; $i++)
    {
      MemTaint($dest, $i);
      SetColor($dest + $i, CIC_ITEM, TaintColor);
    }

    //Display the tainted range
    Message("done\n");
    Message("Tainted buffer: " + sprintf("%08x", $dest));
    ($size == 1) ? Message("\n") : Message(" - " + sprintf("%08x", $dest + $size - 1) + "\n");
  }
  catch($e)
  {
    Message("failed\n");
  }
}

//________________________________________________________________________________________________________________


//Untaint memory by a user
static MemUntaintUser()
{
  //Define variables
  auto $i, $e, $dest, $size;

  //Untaint a buffer
  try
  {
    //Get the memory pointer
    Message("Get: initial untainted memory ...");
    $dest = AskAddr("", "Pointer to memory");
    if (!($dest + 1)) throw 0;
    $size = AskLong("", "Memory size");
    if (!($size+1)) throw 0;

    //Untaint the range
    for ($i=0; $i<$size; $i++)
    {
      MemUntaint($dest, $i);
      SetColor($dest + $i, CIC_ITEM, DEFCOLOR);
    }

    //Display the tainted range
    Message("done\n");
    Message("Untainted buffer: " + sprintf("%08x", $dest));
    ($size == 1) ? Message("\n") : Message(" - " + sprintf("%08x", $dest + $size - 1) + "\n");
  }
  catch($e)
  {
    Message("failed\n");
  }
}

//________________________________________________________________________________________________________________


//Sort and pack the taint memory array
static MemArrayRearrange()
{
  //Define variables
  auto $i, $j, $k;

  //Bubble sort
  for ($j=0; $j<($TaintMemArraySize-1); $j++)
    for ($i=0; $i<($TaintMemArraySize-1-$j); $i++)
      if ($TaintMemArray[$i] > $TaintMemArray[$i+1])
      {
        $k = $TaintMemArray[$i];
	$TaintMemArray[$i] = $TaintMemArray[$i+1];
	$TaintMemArray[$i+1] = $k;
      }

  //Compress the array
  $j = 0;
  for ($i=0; $i<$TaintMemArraySize; $i++)
    if ($TaintMemArray[$i] == Untainted)
      continue;
    else
      $TaintMemArray[$j++] = $TaintMemArray[$i];

  //Set the new size
  $TaintMemArraySize = $j;

  //Exit
  return;
}

//________________________________________________________________________________________________________________


//Display the tainted data
static DisplayTaintData()
{
  //Define variables
  auto $i, $e, $taint;

  //Display tainted registers
  Message("\nTainted registers: ");
  for ($i=0; $i<NN_TMP; $i++)
    try
    {
      $taint = "";
      for ($e=0; $e<RegGetSize($i); $e++)
        if (RegCheck($i, $e))
	  $taint = sprintf("%01x", $e+1) + $taint;
      if ($taint != "") Message(RegGetName($i) + "(" + $taint + ") ");
    }
    catch($e){}

  //Display tainted memory
  Message("\nTainted memory:\n");
  try
  {
    //Sort and pack the taint memory array
    MemArrayRearrange();
    //Display tainted memory regions
    $e = false;
    if ($TaintMemArraySize) Message(sprintf("%08x", $TaintMemArray[0]));
    for ($i=1; $i<$TaintMemArraySize; $i++)
      if (($TaintMemArray[$i-1] + 1) == $TaintMemArray[$i])
        $e = true;
      else
      {
        if ($e)
	{
	  $e = false;
	  Message(" - " + sprintf("%08x", $TaintMemArray[$i-1]));
	}
        Message("\n" + sprintf("%08x", $TaintMemArray[$i]));
      }
    if ($e) Message(" - " + sprintf("%08x", $TaintMemArray[$i-1]));
    Message("\n");
  }
  catch($e){}

  //Exit
  return;
}

//________________________________________________________________________________________________________________


//Callback function to call from the tracing
static TaintSlicer()
{
  //Define variables
  auto $e, $ea, $cmd, $taint;
  auto $TaintComment = " <T>";
  auto $TaintUnsupport = " <U>";

  //Process the trace
  $TracingStop = false;
  try
  {
    //Get the current execution address
    $ea = GetTevEa(TEV_NONE);
    if ($Segment != SegName($ea))
    {
      //For ROP
      $TracingStop = true;
      $CalleeFunc = $ea;
    }

    //Test for an exception
    $cmd = DecodeInstruction($ea);
    if (!$cmd) throw 0;

    //Action of an instruction
    $taint = InstructionAction($ea, $cmd);
    if ($taint == Tainted)
    {
      //Mark the tained instruction
      $e = CommentEx($ea, 0);
      if (strstr($e, $TaintComment) == -1)
      {
	MakeComm($ea, $e + $TaintComment);
	SetColor($ea, CIC_ITEM, TaintColor);
      }
    }
    if ($taint == Unsupported)
    {
      //Mark the unsupported instruction
      $e = CommentEx($ea, 0);
      if (strstr($e, $TaintUnsupport) == -1)
      {
	MakeComm($ea, $e + $TaintUnsupport);
	SetColor($ea, CIC_ITEM, TaintColorUnsup);
      }
    }
  }
  catch($e)
  {
    //Process a error instruction
    Message("Error at: " + sprintf("%08x", $ea) + "\n");
    SetColor($ea, CIC_ITEM, TaintColorErr);
  }

  //Process the stop tracing
  if ($TracingStop)
  {
    //Make a comment with the callee address
    MakeComm($ea, CommentEx($ea, 0) + sprintf("%08x", $CalleeFunc) + " " + Name($CalleeFunc) + "; ");
    DisplayTaintData();
    Message("External code: " + sprintf("%08x", $CalleeFunc) + " " + Name($CalleeFunc) + "\n");
  }

  //Exit
  return $TracingStop;
}

//________________________________________________________________________________________________________________


//The main function
static main()
{
  //Define variables
  auto $i, $e;
  $TaintMemArray = object();
  $TaintRegArray = object();

  //Initialize the array of registers' objects
  $Registers = object();
  $Registers[NN_al  * RegObjectSize + RegName]   = "al";
  $Registers[NN_al  * RegObjectSize + RegOffset] = 0;
  $Registers[NN_al  * RegObjectSize + RegSize]   = 1;
  $Registers[NN_ah  * RegObjectSize + RegName]   = "ah";
  $Registers[NN_ah  * RegObjectSize + RegOffset] = 1;
  $Registers[NN_ah  * RegObjectSize + RegSize]   = 1;
  $Registers[NN_ax  * RegObjectSize + RegName]   = "ax";
  $Registers[NN_ax  * RegObjectSize + RegOffset] = 0;
  $Registers[NN_ax  * RegObjectSize + RegSize]   = 8;
  $Registers[NN_cl  * RegObjectSize + RegName]   = "cl";
  $Registers[NN_cl  * RegObjectSize + RegOffset] = 8;
  $Registers[NN_cl  * RegObjectSize + RegSize]   = 1;
  $Registers[NN_ch  * RegObjectSize + RegName]   = "ch";
  $Registers[NN_ch  * RegObjectSize + RegOffset] = 9;
  $Registers[NN_ch  * RegObjectSize + RegSize]   = 1;
  $Registers[NN_cx  * RegObjectSize + RegName]   = "cx";
  $Registers[NN_cx  * RegObjectSize + RegOffset] = 8;
  $Registers[NN_cx  * RegObjectSize + RegSize]   = 8;
  $Registers[NN_dl  * RegObjectSize + RegName]   = "dl";
  $Registers[NN_dl  * RegObjectSize + RegOffset] = 0x10;
  $Registers[NN_dl  * RegObjectSize + RegSize]   = 1;
  $Registers[NN_dh  * RegObjectSize + RegName]   = "dh";
  $Registers[NN_dh  * RegObjectSize + RegOffset] = 0x11;
  $Registers[NN_dh  * RegObjectSize + RegSize]   = 1;
  $Registers[NN_dx  * RegObjectSize + RegName]   = "dx";
  $Registers[NN_dx  * RegObjectSize + RegOffset] = 0x10;
  $Registers[NN_dx  * RegObjectSize + RegSize]   = 8;
  $Registers[NN_bl  * RegObjectSize + RegName]   = "bl";
  $Registers[NN_bl  * RegObjectSize + RegOffset] = 0x18;
  $Registers[NN_bl  * RegObjectSize + RegSize]   = 1;
  $Registers[NN_bh  * RegObjectSize + RegName]   = "bh";
  $Registers[NN_bh  * RegObjectSize + RegOffset] = 0x19;
  $Registers[NN_bh  * RegObjectSize + RegSize]   = 1;
  $Registers[NN_bx  * RegObjectSize + RegName]   = "bx";
  $Registers[NN_bx  * RegObjectSize + RegOffset] = 0x18;
  $Registers[NN_bx  * RegObjectSize + RegSize]   = 8;
  $Registers[NN_spl * RegObjectSize + RegName]   = "spl";
  $Registers[NN_spl * RegObjectSize + RegOffset] = 0x20;
  $Registers[NN_spl * RegObjectSize + RegSize]   = 1;
  $Registers[NN_sp  * RegObjectSize + RegName]   = "sp";
  $Registers[NN_sp  * RegObjectSize + RegOffset] = 0x20;
  $Registers[NN_sp  * RegObjectSize + RegSize]   = 8;
  $Registers[NN_bpl * RegObjectSize + RegName]   = "bpl";
  $Registers[NN_bpl * RegObjectSize + RegOffset] = 0x28;
  $Registers[NN_bpl * RegObjectSize + RegSize]   = 1;
  $Registers[NN_bp  * RegObjectSize + RegName]   = "bp";
  $Registers[NN_bp  * RegObjectSize + RegOffset] = 0x28;
  $Registers[NN_bp  * RegObjectSize + RegSize]   = 8;
  $Registers[NN_sil * RegObjectSize + RegName]   = "sil";
  $Registers[NN_sil * RegObjectSize + RegOffset] = 0x30;
  $Registers[NN_sil * RegObjectSize + RegSize]   = 1;
  $Registers[NN_si  * RegObjectSize + RegName]   = "si";
  $Registers[NN_si  * RegObjectSize + RegOffset] = 0x30;
  $Registers[NN_si  * RegObjectSize + RegSize]   = 8;
  $Registers[NN_dil * RegObjectSize + RegName]   = "dil";
  $Registers[NN_dil * RegObjectSize + RegOffset] = 0x38;
  $Registers[NN_dil * RegObjectSize + RegSize]   = 1;
  $Registers[NN_di  * RegObjectSize + RegName]   = "di";
  $Registers[NN_di  * RegObjectSize + RegOffset] = 0x38;
  $Registers[NN_di  * RegObjectSize + RegSize]   = 8;
  $Registers[NN_r8  * RegObjectSize + RegName]   = "r8";
  $Registers[NN_r8  * RegObjectSize + RegOffset] = 0x40;
  $Registers[NN_r8  * RegObjectSize + RegSize]   = 8;
  $Registers[NN_r9  * RegObjectSize + RegName]   = "r9";
  $Registers[NN_r9  * RegObjectSize + RegOffset] = 0x48;
  $Registers[NN_r9  * RegObjectSize + RegSize]   = 8;
  $Registers[NN_r10 * RegObjectSize + RegName]   = "r10";
  $Registers[NN_r10 * RegObjectSize + RegOffset] = 0x50;
  $Registers[NN_r10 * RegObjectSize + RegSize]   = 8;
  $Registers[NN_r11 * RegObjectSize + RegName]   = "r11";
  $Registers[NN_r11 * RegObjectSize + RegOffset] = 0x58;
  $Registers[NN_r11 * RegObjectSize + RegSize]   = 8;
  $Registers[NN_r12 * RegObjectSize + RegName]   = "r12";
  $Registers[NN_r12 * RegObjectSize + RegOffset] = 0x60;
  $Registers[NN_r12 * RegObjectSize + RegSize]   = 8;
  $Registers[NN_r13 * RegObjectSize + RegName]   = "r13";
  $Registers[NN_r13 * RegObjectSize + RegOffset] = 0x68;
  $Registers[NN_r13 * RegObjectSize + RegSize]   = 8;
  $Registers[NN_r14 * RegObjectSize + RegName]   = "r14";
  $Registers[NN_r14 * RegObjectSize + RegOffset] = 0x70;
  $Registers[NN_r14 * RegObjectSize + RegSize]   = 8;
  $Registers[NN_r15 * RegObjectSize + RegName]   = "r15";
  $Registers[NN_r15 * RegObjectSize + RegOffset] = 0x78;
  $Registers[NN_r15 * RegObjectSize + RegSize]   = 8;
  $Registers[NN_TMP * RegObjectSize + RegName]   = "TMP";
  $Registers[NN_TMP * RegObjectSize + RegOffset] = 0x80;
  $Registers[NN_TMP * RegObjectSize + RegSize]   = 8;

  //Test for the debugging
  Message("\nTaintSlicer\n");
  Message("Check: debug session ...");
  try
  {
    $Bitness = GetSegmentAttr(GetRegValue("rip"), SEGATTR_BITNESS) + 1;
    $Segment = SegName(GetRegValue("rip"));
  }
  catch($e)
  {
    try
    {
      $Bitness = GetSegmentAttr(GetRegValue("eip"), SEGATTR_BITNESS);
      $Segment = SegName(GetRegValue("eip"));
    }
    catch($e)
    {
      Message("failed\n");
      return;
    }
  }
  Message("done\n");
  Message("Current segment: " + $Segment + "\n");
  Message("Bitness: ");
  ($Bitness == 1) ? Message("32\n") : Message("64\n");

  //Clear the array of tainted registers and memory
  $TaintMemArraySize = 0;
  for ($i=0; $i<(NN_all*8); $i++)
    $TaintRegArray[$i] = Untainted;

  //Setup hot keys
  AddHotkey("Alt-Ctrl-M", "MemTaintUser");
  AddHotkey("Ctrl-M",     "MemUntaintUser");
  AddHotkey("Alt-Ctrl-R", "RegTaintUser");
  AddHotkey("Ctrl-R",     "RegUntaintUser");
  AddHotkey("Alt-Ctrl-T", "DisplayTaintData");

  //Display settings for the tracing
  $e = "Taint memory: Alt-Ctrl-M\n" +
       "Untaint memory: Ctrl-U\n" +
       "Taint registers: Alt-Ctrl-R\n" +
       "Untaint registers: Ctrl-R\n" +
       "Display tainted data: Alt-Ctrl-T\n" +
       "Set Tracing options:\n" +
       "- Trace buffer size: 0\n" +
       "- Stop condition: TaintSlicer()\n" +
       "- Trace over debugger segments\n" +
       "- Instruction tracing\n";
  Message($e);
}

//________________________________________________________________________________________________________________

